import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import copy
from collections import defaultdict

# creates the duplicate index
def create_dupe_index(df, ind_name):
	df[ind_name+'_back'] = df[ind_name]
	return df

# renames old index dupe column in df and resets the index
def reset_dupe_index(df, ind_name):
	df.rename({ind_name: ind_name+'_back'}, inplace=True, axis=1)
	df.reset_index(inplace=True)
	return(df)

# set index, rename dupe index in df
def set_dupe_index(df, ind_name):
	df.set_index(ind_name, inplace=True)
	df.rename({ind_name+'_back': ind_name}, inplace=True, axis=1)
	return(df)

# partner function to label_edges
def set_edge_attrs(x, G, f_df, f_e):
	attr = {(x.v1, x.v2): {f_e: x[f_df]}}
	nx.set_edge_attributes(G, attr)
	return G

# label edges in G based on fields of edge_df
def label_edges(G, edge_df, f_df, f_e):
	edge_df.apply(lambda x: set_edge_attrs(x, G, f_df, f_e), axis=1)
	return G

# parter function to label_nodes
def set_node_attrs(x, G, f_df, f_n):
	attr = {x.vertex_id: {f_n: x[f_df]}}
	nx.set_node_attributes(G, attr)
	return G

# label nodes in G based on fields of loc_df
def label_nodes(G, loc_df, f_df, f_n):
	loc_df.apply(lambda x: set_node_attrs(x, G, f_df, f_n), axis=1)
	return G

# get value associated with keyword in the 9th column of gtf
def get_field_value(key, fields):
    if key not in fields:
        return None
    else:
        return fields.split(key+' "')[1].split()[0].replace('";','')

# creates a dictionary of the last field of a gtf
# adapted from Dana Wyman
def get_fields(fields):

    attributes = {}

    description = fields.strip()
    description = [x.strip() for x in description.split(";")]
    for pair in description:
        if pair == "": continue

        pair = pair.replace('"', '')
        key, val = pair.split()
        attributes[key] = val

    # put in placeholders for important attributes (such as gene_id) if they
    # are absent
    if "gene_id" not in attributes:
        attributes["gene_id"] = "NULL"

    return attributes    

# return a table indexed by transcript id with the appropriate 
# abundance
# currently only works with TALON abundance files but can easily 
# be updated to work with more types of abundance files
def process_abundance_file(file, cols):

	if type(cols) != list: cols = [cols]

	df = pd.read_csv(file, sep='\t')
	keep_cols = ['annot_transcript_id']+cols	
	df = df[keep_cols]

	# get the counts
	df['counts'] = df[cols].sum(axis=1)

	# get tpms
	for col in cols: 
		total_counts = df[col].sum()
		df['{}_tpm'.format(col)] = (df[col]*1000000)/total_counts
	cols = ['{}_tpm'.format(col) for col in cols]
	df['tpm'] = df[cols].mean(axis=1)

	# set up for merging
	df.drop(cols, axis=1, inplace=True)
	df.rename({'annot_transcript_id': 'tid'}, inplace=True, axis=1)

	return df

# creates a file name based on input plotting arguments
def create_fname(prefix, combine, indicate_dataset,
				 indicate_novel, browser,
				 ftype='figure', tid=None, gid=None):
	fname = prefix
	if combine:
		fname += '_combine'
	if indicate_dataset:
		fname += '_{}'.format(indicate_dataset)
	if indicate_novel:
		fname += '_novel'
	if browser: 
		fname += '_browser'
	if tid: 
		fname += '_{}'.format(tid)
	if gid: 
		fname += '_{}'.format(gid)
	if ftype == 'figure':
		fname += '.png'
	elif ftype == 'report':
		fname += '_report.pdf'
	return fname

# checks if a file is a gtf or a db
def gtf_or_db(fname):
	ext = fname.split('.')[-1]
	if ext == 'gtf': return 'gtf'
	elif ext == 'db': return 'db'
	else:
		raise Exception('File type must be gtf or db. '
			'Type received is {}'.format(ext))



	









import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import math
import copy
from collections import defaultdict
import sqlite3
import pickle
from swan.utils import *
from swan.graph import Graph
from swan.plottedgraph import PlottedGraph

class SwanGraph(Graph):

	def __init__(self):
		super().__init__()

	###########################################################################
	############## Related to adding datasets and merging #####################
	###########################################################################

	# add annotation to graph 
	def add_annotation(self, fname):

		# column name for annotation 
		col = 'annotation'

		# use the add_dataset function to add stuff to graph
		self.add_dataset(col, fname)

	# add dataset into graph from gtf
	def add_dataset(self, col, fname, dname=None,
					counts_file=None, count_cols=None):

		# make sure that input dataset name is not
		# already in any of the df col spaces
		if col in self.datasets:
			raise Exception('Dataset {} is already in the graph. '
				'Use update_dataset (coming soon) or provide a different name.'.format(col))
		if col in self.loc_df.columns:
			raise Exception('Dataset name {} conflicts with preexisting '
				'column in loc_df. Choose a different name.'.format(col))
		if col in self.edge_df.columns:
			raise Exception('Dataset name {} conflicts with preexisting '
				'column in edge_df. Choose a different name.'.format(col))
		if col in self.t_df.columns:
			raise Exception('Dataset name {} conflicts with preexisting '
				'column in t_df. Choose a different name.'.format(col))

		# are we dealing with a gtf or a db?
		ftype = gtf_or_db(fname)

		# first entry is easy 
		if self.is_empty():

			# get loc_df, edge_df, t_df
			if ftype == 'gtf':
				self.create_dfs_gtf(fname)
			elif ftype == 'db':
				self.create_dfs_db(fname, dname)

			# add column to each df to indicate where data came from
			self.loc_df[col] = True
			self.edge_df[col] = True
			self.t_df[col] = True

		# adding a new dataset to the graph requires us to merge
		# SwanGraph objects
		else:
			temp = SwanGraph()
			if ftype == 'gtf':
				temp.create_dfs_gtf(fname)
			elif ftype == 'db':
				temp.create_dfs_db(fname, dname)
			self.merge_dfs(temp, col)

		# order node ids by genomic position, add node types,
		# and create graph
		self.update_ids()
		self.order_edge_df()
		self.get_loc_types()
		self.create_graph_from_dfs()

		# update graph metadata
		self.datasets.append(col)

		# if we're also adding abundances
		if counts_file and count_cols:
			self.add_abundance(counts_file, count_cols, col)

	# adds counts columns to t_df based on columns counts_cols found in 
	# tsv counts_file. Relies on column annot_transcript_id to have the
	# transcript id (tid) that is used to index t_df. 
	# TODO make this more flexible in the future
	def add_abundance(self, counts_file, count_cols, dataset_name):

		# if the dataset we're trying to add counts too doesn't exist
		if dataset_name not in self.datasets:
			raise Exception('Trying to add expression data to a dataset '
							'that is not in the graph. Add dataset to graph first.')

		# get counts from input abundance file 
		abundance_df = process_abundance_file(counts_file, count_cols)
		abundance_df.rename({'tpm': '{}_tpm'.format(dataset_name),
							 'counts': '{}_counts'.format(dataset_name)},
							 axis=1, inplace=True)

		# merge on transcript id (tid) with t_df and make sure it's 
		# formatted correctly
		self.t_df.reset_index(drop=True, inplace=True)
		self.t_df = self.t_df.merge(abundance_df, on='tid', how='left')
		self.t_df.fillna(value=0, inplace=True)
		self.t_df = create_dupe_index(self.t_df, 'tid')
		self.t_df = set_dupe_index(self.t_df, 'tid')

		# finally update object's metadata
		self.counts.append('{}_counts'.format(dataset_name))
		self.tpm.append('{}_tpm'.format(dataset_name))

	# merge dfs from two SwanGraph objects
	def merge_dfs(self, b, b_col):

		# merge loc dfs
		# what locations correspond between the datasets?
		self.merge_loc_dfs(b, b_col)
		id_map = self.get_merged_id_map()

		self.loc_df.drop(['vertex_id_a','vertex_id_b'], axis=1, inplace=True)
		self.loc_df['vertex_id'] = self.loc_df.index
		self.loc_df = create_dupe_index(self.loc_df, 'vertex_id')
		self.loc_df = set_dupe_index(self.loc_df, 'vertex_id')
		b.loc_df = create_dupe_index(b.loc_df, 'vertex_id')
		b.loc_df = set_dupe_index(b.loc_df, 'vertex_id')

		# update the ids in b to make edge_df, t_df merging easier
		b.update_ids(id_map=id_map)

		# merge edge_df and t_df
		self.merge_edge_dfs(b, b_col)
		self.merge_t_dfs(b, b_col)

	# merge t_dfs on tid, gid, gname, path
	def merge_t_dfs(self, b, b_col):

		# some df reformatting
		self.t_df.reset_index(drop=True, inplace=True)
		b.t_df.reset_index(drop=True, inplace=True)
		b.t_df[b_col] = True

		# convert paths to tuples so we can merge on them
		self.t_df.path = self.t_df.apply(
			lambda x: tuple(x.path), axis=1)
		b.t_df.path = b.t_df.apply(
			lambda x: tuple(x.path), axis=1)

		# merge on transcript information
		t_df = self.t_df.merge(b.t_df,
			   how='outer',
			   on=['tid', 'gid', 'gname', 'path'],
			   suffixes=['_a', '_b'])

		# convert path back to list
		t_df.path = t_df.apply(
			lambda x: list(x.path), axis=1)

		# assign False to entries that are not in the new dataset, 
		# and to new entries that were not in the prior datasets
		d_cols = self.datasets+[b_col]
		t_df[d_cols] = t_df[d_cols].fillna(value=False, axis=1)

		# set up index again
		t_df = create_dupe_index(t_df, 'tid')
		t_df = set_dupe_index(t_df, 'tid')

		self.t_df = t_df

	# merge edge_dfs on edge_id, v1, v2, strand, edge_type
	def merge_edge_dfs(self, b, b_col):

		# some df reformatting
		self.edge_df.reset_index(drop=True, inplace=True)
		b.edge_df.reset_index(drop=True, inplace=True)
		b.edge_df[b_col] = True

		# merge on edge info
		edge_df = self.edge_df.merge(b.edge_df,
				  how='outer',
				  on=['edge_id', 'v1', 'v2', 'edge_type', 'strand'],
				  suffixes=['_a', '_b'])

		# assign False to entries that are not in the new dataset, 
		# and to new entries that were not in the prior datasets
		d_cols = self.datasets+[b_col]
		edge_df[d_cols] = edge_df[d_cols].fillna(value=False, axis=1)

		# remake index
		edge_df = create_dupe_index(edge_df, 'edge_id')
		edge_df = set_dupe_index(edge_df, 'edge_id')
		
		self.edge_df = edge_df

	# merge loc_dfs on coord, chrom, strand
	def merge_loc_dfs(self, b, b_col):

		# some df reformatting
		node_types = ['TSS', 'alt_TSS', 'TES', 'alt_TES', 'internal']

		self.loc_df.drop(node_types, axis=1, inplace=True)
		self.loc_df.reset_index(drop=True, inplace=True)

		# b.loc_df.drop(node_types, axis=1, inplace=True)
		b.loc_df.reset_index(drop=True, inplace=True)
		b.loc_df[b_col] = True

		# merge on location info
		loc_df = self.loc_df.merge(b.loc_df,
				 how='outer',
				 on=['chrom', 'coord', 'strand'],
				 suffixes=['_a','_b'])

		# assign False to entries that are not in the new dataset, 
		# and to new entries that were not in prior datasets
		d_cols = self.datasets+[b_col]
		loc_df[d_cols] = loc_df[d_cols].fillna(value=False, axis=1)

		self.loc_df = loc_df

	# returns a dictionary mapping vertex b: vertex a for each
	# vertex in dataset b
	def get_merged_id_map(self):

		id_map = self.loc_df.apply(
			lambda x: [x.vertex_id_b, x.vertex_id_a], axis=1)

		# loop through id_map and assign new ids for 
		# those present in b but not a
		b_ind = int(self.loc_df.vertex_id_a.max() + 1)
		i = 0
		while i < len(id_map):
			if math.isnan(id_map[i][1]):
				id_map[i][1] = b_ind
				b_ind += 1
			# set up entries where there isn't a b id (entries only found in a)
			# to be removed
			elif math.isnan(id_map[i][0]):
				id_map[i] = []
			i += 1

		# remove entries that are only in a but not in b
		# make sure everything is ints
		id_map = [i for i in id_map if len(i) == 2]
		id_map = dict([(int(a), int(b)) for a,b in id_map])

		return id_map

	##########################################################################
	############# Related to creating dfs from GTF or TALON DB ###############
	##########################################################################

	# create loc_df (nodes), edge_df (edges), and t_df (transcripts) from gtf
	# adapted from Dana Wyman and TALON
	def create_dfs_gtf(self, gtf_file):

		# make sure file exists
		if not os.path.exists(gtf_file):
			raise Exception('GTF file not found. Check path.')

		# depending on the strand, determine the stard and stop
		# coords of an intron or exon
		def find_edge_start_stop(v1, v2, strand):
			if strand == '-':
				start = max([v1, v2])
				stop = min([v1, v2])
			elif strand == '+':
				start = min([v1, v2])
				stop = max([v1, v2])
			return start, stop

		# dictionaries to hold unique edges and transcripts
		transcripts = {}
		exons = {}

		with open(gtf_file) as gtf:
			for line in gtf:

				# ignore header lines
				if '##' in line:
					continue

				# split each entry
				line = line.strip().split('\t')

				# get some fields from gtf that we care about
				chrom = line[0]
				entry_type = line[2]
				start = int(line[3])
				stop = int(line[4])
				strand = line[6]
				fields = line[-1]

				# transcript entry 
				if entry_type == "transcript":
					attributes = get_fields(fields)
					tid = attributes['transcript_id']
					gid = attributes['gene_id']
					gname = attributes['gene_name']

					# add transcript to dictionary 
					transcript = {tid: {'gid': gid,
										'gname': gname,
										'tid': tid,
										'strand': strand,
										'exons': []}}
					transcripts.update(transcript)
					
				# exon entry
				elif entry_type == "exon":
					attributes = get_fields(fields)
					start, stop = find_edge_start_stop(start, stop, strand)
					eid = '{}_{}_{}_{}_exon'.format(chrom, start, stop, strand)
					tid = attributes['transcript_id']	

					# add novel exon to dictionary 
					if eid not in exons:
						edge = {eid: {'eid': eid,
									  'chrom': chrom,
									  'v1': start,
									  'v2': stop,
									  'strand': strand}}
						exons.update(edge)
			   
			   		# add this exon to the transcript's list of exons
					if tid in transcripts:
						transcripts[tid]['exons'].append(eid)

		# once we have all transcripts, make loc_df
		locs = {}
		vertex_id = 0
		for edge_id, edge in exons.items():
			chrom = edge['chrom']
			strand = edge['strand']

			v1 = edge['v1']
			v2 = edge['v2']

			# exon start
			key = (chrom, v1, strand)
			if key not in locs:
				locs[key] = vertex_id
				vertex_id += 1
			# exon end
			key = (chrom, v2, strand)
			if key not in locs:
				locs[key] = vertex_id
				vertex_id += 1

		# add locs-indexed path to transcripts, and populate edges
		edges = {}
		for _,t in transcripts.items():
			t['path'] = []
			strand = t['strand']
			t_exons = t['exons']

			for i, exon_id in enumerate(t_exons):

				# pull some information from exon dict
				exon = exons[exon_id]
				chrom = exon['chrom']
				v1 = exon['v1']
				v2 = exon['v2']
				strand = exon['strand']

				# add current exon and subsequent intron 
				# (if not the last exon) for each exon to edges
				key = (chrom, v1, v2, strand)
				v1_key = (chrom, v1, strand)
				v2_key = (chrom, v2, strand)
				edge_id = (locs[v1_key], locs[v2_key])
				if key not in edges:
					edges[key] = {'edge_id': edge_id, 'edge_type': 'exon'}

				# add exon locs to path
				t['path'] += list(edge_id)

				# if this isn't the last exon, we also needa add an intron
				# this consists of v2 of the prev exon and v1 of the next exon
				if i < len(t_exons)-1:
					next_exon = exons[t_exons[i+1]]
					v1 = next_exon['v1']
					key = (chrom, v2, v1, strand)
					v1_key = (chrom, v1, strand)
					edge_id = (locs[v2_key], locs[v1_key])
					if key not in edges:
						edges[key] = {'edge_id': edge_id, 'edge_type': 'intron'}

		# turn transcripts, edges, and locs into dataframes
		locs = [{'chrom': key[0],
				 'coord': key[1],
				 'strand': key[2],
				 'vertex_id': vertex_id} for key, vertex_id in locs.items()]
		loc_df = pd.DataFrame(locs)

		edges = [{'v1': item['edge_id'][0],
				  'v2': item['edge_id'][1], 
				  'strand': key[3],
				  'edge_id': item['edge_id'],
				  'edge_type': item['edge_type']} for key, item in edges.items()]
		edge_df = pd.DataFrame(edges)

		transcripts = [{'tid': key,
						'gid': item['gid'],
						'gname': item['gname'],
						'path': item['path']} for key, item in transcripts.items()]
		t_df = pd.DataFrame(transcripts)

		# final df formatting
		loc_df = create_dupe_index(loc_df, 'vertex_id')
		loc_df = set_dupe_index(loc_df, 'vertex_id')
		edge_df = create_dupe_index(edge_df, 'edge_id')
		edge_df = set_dupe_index(edge_df, 'edge_id')
		t_df = create_dupe_index(t_df, 'tid')
		t_df = set_dupe_index(t_df, 'tid')

		self.loc_df = loc_df
		self.edge_df = edge_df
		self.t_df = t_df

	# create loc_df (for nodes), edge_df (for edges), and t_df (for paths)
	def create_dfs_db(self, db, dname):

		# are we pulling a specific dataset's observed transcripts
		# from the database?
		if dname == None:
			print('No dataset name given.'
				' Will pull everything, including unobserved transcripts,'
				' from {}'.format(db))
		else:
			print('Getting transcripts for {} from {}'.format(db, dname))

		# make sure file exists
		if not os.path.exists(db):
			raise Exception('TALON db file {} not found. Check path.'.format(db))

		# open db connection
		conn = sqlite3.connect(db)
		c = conn.cursor()

		# t_df
		t_df = pd.DataFrame()

		# get tid, gid, gname, and paths
		q = """SELECT ga.value, ta.value,
					  t.start_exon, t.jn_path, t.end_exon,
					  t.start_vertex, t.end_vertex
				FROM gene_annotations ga 
				JOIN transcripts t ON ga.ID=t.gene_ID
				JOIN transcript_annotations ta ON t.transcript_ID=ta.ID
			"""
		if dname: 
			q += """ JOIN observed o ON o.transcript_ID=t.transcript_ID"""
		q += """ WHERE ta.attribute='transcript_id'
				AND (ga.attribute='gene_name' 
				OR ga.attribute='gene_id')
			"""
		if dname: 
			q += """ AND o.dataset='{}'""".format(dname)

		c.execute(q)
		data = c.fetchall()

		# get fields from each transcript and add to dataframe
		gids, tids, paths = zip(*[(i[0], i[1], i[2:]) for i in data[::2]])
		gnames = [i[0] for i in data[1::2]]
		paths = self.get_db_edge_paths(paths)

		t_df['tid'] = np.asarray(tids)
		t_df['gid'] = np.asarray(gids)
		t_df['gname'] = np.asarray(gnames)
		t_df['path'] = np.asarray(paths)

		t_df = create_dupe_index(t_df, 'tid')
		t_df = set_dupe_index(t_df, 'tid')

		# edge_df

		# get the list of edge ids we need to pull from db
		edge_ids = list(set([str(n) for path in paths for n in path]))
		edge_str = '({})'.format(','.join(edge_ids))

		q = """SELECT DISTINCT e.* 
				FROM edge e 
				JOIN vertex V ON e.v1=v.vertex_ID 
			"""
		if dname:
			q += """ WHERE e.edge_ID IN {}""".format(edge_str)

		c.execute(q)
		edges = c.fetchall()

		edge_df = pd.DataFrame(edges, 
			columns=['edge_id', 'v1', 'v2',
					 'edge_type', 'strand'])
		edge_df.v1 = edge_df.v1.map(int)
		edge_df.v2 = edge_df.v2.map(int)
		edge_df['talon_edge_id'] = edge_df.edge_id
		edge_df['edge_id'] = edge_df.apply(lambda x: (int(x.v1), int(x.v2)), axis=1)

		# loc_df

		# get the list of vertex ids we need to pull from db
		edge_ids = edge_df.edge_id.tolist()
		loc_ids = list(set([str(n) for edge_id in edge_ids for n in edge_id]))
		loc_string = '({})'.format(','.join(loc_ids))


		q = """SELECT loc.* FROM location loc"""
		if dname:
			q += """ WHERE loc.location_ID IN {}""".format(loc_string)

		c.execute(q)
		locs = c.fetchall()

		loc_df = pd.DataFrame(locs,
			columns=['location_ID', 'genome_build',
					 'chrom', 'position'])

		# do some df reformatting, add strand
		loc_df.drop('genome_build', axis=1, inplace=True)
		loc_df.rename({'location_ID': 'vertex_id',
					   'position': 'coord'},
					   inplace=True, axis=1)
		loc_df.vertex_id = loc_df.vertex_id.map(int)

		# furnish the last bit of info in each df
		t_df['path'] = [[int(n) for n in path]
						 for path in self.get_db_vertex_paths(paths, edge_df)]

		# get strandedness of each sj
		extra_entries = pd.DataFrame(columns=['vertex_id', 'chrom',
			'coord', 'strand', 'plus_vertex_id'])
		loc_df['strand'] = np.nan
		for index, entry in loc_df.iterrows():

			# use v1 or v2 depending on where vertex is in edge
			loc_id = entry.vertex_id
			strands = edge_df.loc[(edge_df.v1==loc_id) | (edge_df.v2==loc_id), 'strand']
			strands = list(set(strands.tolist()))

			# if this sj is found on both the + and - strand,
			# create a minus strand entry to add after the apply
			# function is done. return the plus strand entry
			if len(strands) > 1:
				minus_entry = entry.copy(deep=True)
				minus_entry.strand = '-'
				minus_entry.vertex_id = len(loc_df.index)+len(extra_entries.index)+1
				minus_entry['plus_vertex_id'] = entry.vertex_id
				extra_entries = extra_entries.append(minus_entry)

				loc_df.loc[index, 'strand'] = '+'
			else:
				loc_df.loc[index, 'strand'] = strands[0]
		loc_df = loc_df.append(extra_entries, ignore_index=True, sort=True)

		loc_df = create_dupe_index(loc_df, 'vertex_id')
		loc_df = set_dupe_index(loc_df, 'vertex_id')

		# replace sjs in t_df paths that had to be introduced in extra entries
		double_sjs = extra_entries.plus_vertex_id.tolist()
		for tid, entry in t_df.iterrows():
			path = entry.path
			if any(x in path for x in double_sjs):
				coords = loc_df.loc[path[:2], 'coord'].tolist()

				# this transcript is on the minus strand
				if coords[1] < coords[0]:
					new_path = [extra_entries.loc[extra_entries.plus_vertex_id == n,
						'vertex_id'].tolist()[0] if n in double_sjs else n for n in path]
					t_df.loc[tid, 'path'] = new_path

		# also replace sjs in edge_df that had to be introduced in extra_entries
		edge_df.drop('talon_edge_id', axis=1, inplace=True)
		edge_df = create_dupe_index(edge_df, 'edge_id')
		edge_df = set_dupe_index(edge_df, 'edge_id')
		edge_df.to_csv('input_files/edge_df_db.csv')
		for eid, entry in edge_df.iterrows():
			if any(x in eid for x in double_sjs):
				coords = loc_df.loc[eid, 'coord'].tolist()	

				# this edge is on the minus strand
				if coords[1] < coords[0]:
					new_eid = tuple([extra_entries.loc[extra_entries.plus_vertex_id == n, 'vertex_id'].tolist()[0] if n in double_sjs else n for n in eid])
					edge_df.loc[[eid], 'v1'] = new_eid[0]
					edge_df.loc[[eid], 'v2'] = new_eid[1]

		# reset edge ids to use the ones updated from extra_entries
		edge_df.reset_index(drop=True, inplace=True)
		edge_df['edge_id'] = edge_df.apply(lambda x: (int(x.v1), int(x.v2)), axis=1)
		edge_df = create_dupe_index(edge_df, 'edge_id')
		edge_df = set_dupe_index(edge_df, 'edge_id')

		# finally, drop unused column from loc_df
		loc_df.drop('plus_vertex_id', axis=1, inplace=True)

		self.loc_df = loc_df
		self.edge_df = edge_df
		self.t_df = t_df

	# add node types (internal, TSS, alt TSS, TES, alt_TES) to loc_df
	def get_loc_types(self):

		def label_node_types(locs, vertex_ids, node_type):
			for vertex_id in vertex_ids:
				locs[vertex_id][node_type] = True
			return locs 

		loc_df = self.loc_df
		t_df = self.t_df

		# create a dictionary to hold loc info to speed stuff up
		locs = loc_df.loc[:, ['vertex_id', 'coord']].copy(deep=True)
		locs['internal'] = False
		locs['TSS'] = False
		locs['TES'] = False
		locs['alt_TSS'] = False
		locs['alt_TES'] = False
		locs.drop(['coord', 'vertex_id'], axis=1, inplace=True)
		locs = locs.to_dict('index')

		# label each TSS and TES
		paths = t_df.path.tolist()
		tss = np.unique([path[0] for path in paths])
		locs = label_node_types(locs, tss, 'TSS')
		tes = np.unique([path[-1] for path in paths])
		locs = label_node_types(locs, tes, 'TES')
		internal = np.unique([n for path in paths for n in path[1:-1]])
		locs = label_node_types(locs, internal, 'internal')

		# also create a dictionary of gid: [path1, ... pathn] to speed up
		path_dict = defaultdict(list)
		for tid, entry in t_df.iterrows():
			path_dict[entry.gid].append(entry.path)

		# label alt TES/TSS
		for gid in path_dict.keys():
			paths = path_dict[gid]
			if len(paths) > 1:
				tss = np.unique([path[0] for path in paths])
				tes = np.unique([path[-1] for path in paths])
				if len(tss) > 1:  
					locs = label_node_types(locs, tss, 'alt_TSS')
				if len(tes) > 1:
					locs = label_node_types(locs, tes, 'alt_TES')

		# create df from locs dict and append old loc_df with node types
		locs = pd.DataFrame.from_dict(locs, orient='index')
		loc_df = pd.concat([loc_df, locs], axis=1)
		loc_df = create_dupe_index(loc_df, 'vertex_id')
		loc_df = set_dupe_index(loc_df, 'vertex_id')

		self.loc_df = loc_df
		self.t_df = t_df

	# convert talon query into edge path
	def get_db_edge_paths(self, paths):
		edge_paths = []
		for p in paths:
			if p[1] == None:
				edge_paths.append([p[0]])
			else:
				edge_paths.append(
					[p[0], *[int(i) for i in p[1].split(',')], p[2]])
		return edge_paths

	# convert edge path to vertex path
	def get_db_vertex_paths(self, paths, edge_df):
		vertex_paths = []
		for p in paths: 
			path = []
			for i, e in enumerate(p): 
				entry = edge_df.loc[edge_df.talon_edge_id == e]
				if i == 0:
					path.extend([entry.v1.values[0], entry.v2.values[0]])
				else: path.append(entry.v2.values[0])
			vertex_paths.append(path)
		return vertex_paths

	##########################################################################
	######################## Other SwanGraph utilities #####################
	##########################################################################

	# order the transcripts by expression of transcript, transcript id, 
	# or start/end nodes
	def order_transcripts(self, order='tid'):

		# order by transcript id
		if order == 'tid':
			ordered_tids = sorted(self.t_df.tid.tolist())
			self.t_df = self.t_df.loc[ordered_tids]

		# order by expression
		elif order == 'expression':
			tpm_cols = self.get_tpm_cols()

			# make sure there are counts in the graph at all
			if tpm_cols:
				self.t_df['tpm_sum'] = self.t_df[tpm_cols].sum(axis=1)
				self.t_df.sort_values(by='tpm_sum', 
									  ascending=False, 
									  inplace=True)
				self.t_df.drop('tpm_sum', axis=1, inplace=True)
			else: 
				raise Exception('Cannot order by expression because '
								'there is no expression data.')

		# order by coordinate of tss
		# TODO might be able to roll this in with getting ordered_nodes
		# in PlottedGraph
		elif order == 'tss':
			self.t_df['start_coord'] = self.t_df.apply(lambda x: 
				self.loc_df.loc[x.path[0], 'coord'], axis=1)

			# watch out for strandedness
			if self.loc_df.loc[self.loc_df.index[0], 'strand'] == '-':
				ascending = False
			else: 
				ascending = True
			self.t_df.sort_values(by='start_coord',
								  ascending=ascending,
								  inplace=True)
			self.t_df.drop('start_coord', axis=1, inplace=True)
			
		# order by coordinate of tes
		elif order == 'tes':
			self.t_df['end_coord'] = self.t_df.apply(lambda x: 
				self.loc_df.loc[x.path[-1], 'coord'], axis=1)

			# watch out for strandedness
			if self.loc_df.loc[self.loc_df.index[0], 'strand'] == '-':
				ascending = False
			else: 
				ascending = True
			self.t_df.sort_values(by='end_coord',
								  ascending=ascending,
								  inplace=True)
			self.t_df.drop('end_coord', axis=1, inplace=True)

	##########################################################################
	######################## Finding "interesting" genes #####################
	##########################################################################

	# returns a list of genes that are "interesting"
	def find_genes_with_novel_isoforms(self):

		# get all the datasets, make sure we're not counting transcripts 
		# that are only in the annotation
		if 'annotation' not in self.datasets:
			raise Exception('No annotation data in graph. Cannot ',
				'determine isoform novelty.')
		datasets = self.get_dataset_cols(include_annotation=False)
		t_df = self.t_df.copy(deep=True)
		t_df = t_df.loc[t_df[datasets].any(axis=1)]

		# how many known and novel isoforms does each gene have
		t_df['known'] = t_df.annotation
		t_df['novel'] = [not i for i in t_df.annotation.tolist()]
		keep_cols = ['annotation', 'known', 'novel', 'gid']
		g_df = t_df[keep_cols].groupby(['gid']).sum()

		# create 'interestingness' column ranking how many novel 
		# compared to known isoforms there are, also ranked by 
		# number of total isoforms
		g_df.known = g_df.known.astype('int32')
		g_df.novel = g_df.novel.astype('int32')
		g_df['interestingness'] = ((g_df.novel+1)/(g_df.known+1))*(g_df.known+g_df.novel)
		g_df.sort_values(by='interestingness', ascending=False, inplace=True)

		# top 10 in case the user doesn't care about whole df
		genes = g_df.index.tolist()[:10]

		return genes, g_df

	# TODO find genes with higher expression in novel than known isoforms
	def find_genes_with_high_novel_expression():
		pass

	# find differentially expressed genes
	# d1 = list or string containing first set of datasets to analyze
	# d2 = list or string containing second set of datasets to analyze
	# the above two will be compared for differential experssion
	def find_differentially_expressed_genes(self, d1, d2):

		# first check to see if the queried expression data is in the graph
		if type(d1) != list: d1 = [d1]
		if type(d2) != list: d2 = [d2]
		datasets = d1+d2
		self.check_abundances(datasets)

		# group t_df into gene df and sum up abundances
		# both across genes and across datasets
		t_df = self.t_df.copy(deep=True)
		d1_cols = self.get_tpm_cols(d1)
		d2_cols = self.get_tpm_cols(d2)
		keep_cols = d1_cols+d2_cols+['gid']
		g_df = t_df[keep_cols].groupby('gid').sum()
		g_df['d1_tpm'] = g_df[d1_cols].sum(axis=1)
		g_df['d2_tpm'] = g_df[d2_cols].sum(axis=1)

		# add pseudocounts for each gene
		g_df.d1_tpm = g_df.d1_tpm + 1
		g_df.d2_tpm = g_df.d2_tpm + 1

		# calculate log2 fold change and order g_df based on magnitude
		# of fold change
		g_df['log_fold_change'] = np.log2(g_df.d1_tpm/g_df.d2_tpm)
		g_df['abs_log_fold_change'] = np.absolute(g_df.log_fold_change)
		g_df.sort_values(by='abs_log_fold_change', ascending=False, inplace=True)

		# top 10 in case the user doesn't care about whole df
		genes = g_df.index.tolist()[:10]

		return genes, g_df

	# find differentially expressed transcripts
	def find_differentially_expressed_transcripts():
		pass

	##########################################################################
	######################## Loading/saving SwanGraphs #####################
	##########################################################################

	# saves a splice graph object in pickle format
	def save_graph(self, prefix):
		picklefile = open(prefix+'.p', 'wb')
		pickle.dump(self, picklefile)
		picklefile.close()

	# loads a splice graph object from pickle form
	def load_graph(self, fname):

		picklefile = open(fname, 'rb')
		graph = pickle.load(picklefile)

		# assign SwanGraph fields from file to self
		self.loc_df = graph.loc_df
		self.edge_df = graph.edge_df
		self.t_df = graph.t_df
		self.datasets = graph.datasets
		self.counts = graph.counts
		self.tpm = graph.tpm
		self.pg = graph.pg
		self.G = graph.G

		picklefile.close()

		print('Graph from {} loaded'.format(fname))

	##########################################################################
	############################ Plotting utilities ##########################
	##########################################################################


	# plot the SwanGraph object according to the user's input
	def plot_graph(self, gid, combine=False,
				   indicate_dataset=False,
				   indicate_novel=False):

		self.check_plotting_args(combine, indicate_dataset, indicate_novel)

		# TODO check to see if we already have a pg object and 
		# reinitialize based on compatibility of object 

		# create PlottedGraph object and plot summary graph
		self.pg = PlottedGraph(self, combine, indicate_dataset, indicate_novel, gid=gid)
		self.pg.plot_graph()


	# plot an input transcript's path through the summary graph 
	def plot_transcript_path(self, tid, combine=False,
							 indicate_dataset=False,
							 indicate_novel=False,
							 browser=False):

		self.check_plotting_args(combine, indicate_dataset, indicate_novel, browser)

		# create PlottedGraph object
		self.pg = PlottedGraph(self,
							   combine,
							   indicate_dataset,
							   indicate_novel,
							   tid=tid,
							   browser=browser)

		self.pg.plot_graph()

	# plots each transcript's path through the summary graph, and automatically saves them!
	def plot_each_transcript(self, gid, prefix, combine=False,
							 indicate_dataset=False,
							 indicate_novel=False,
							 browser=False):

		self.check_plotting_args(combine, indicate_dataset, indicate_novel, browser)

		# loop through each transcript in the SwanGraph object
		for tid in self.t_df.loc[self.t_df.gid == gid, 'tid'].tolist():
			self.pg = PlottedGraph(self,
								   combine,
								   indicate_dataset,
								   indicate_novel,
								   tid=tid,
								   gid=gid,
								   browser=browser)
			fname = create_fname(prefix,
								 combine,
								 indicate_dataset,
								 indicate_novel,
								 browser,
								 tid=tid)
			self.pg.plot_graph()
			print('Saving plot for {} as {}'.format(tid, fname))
			self.save_fig(fname)

	# saves current figure named oname. clears the figure space so additional
	# plotting can be done
	def save_fig(self, oname):
		plt.axis('off')
		plt.tight_layout()
		plt.savefig(oname, format='png', dpi=200)
		plt.clf()
		plt.close()

	##########################################################################
	############################### Report stuff #############################
	##########################################################################
	# creates a report for each transcript model for a gene according to user input
	def gen_report(self,
				   gids,
				   prefix,
				   datasets='all',
				   heatmap=False,
				   tpm=False,
				   include_unexpressed=False,
				   combine=False,
				   indicate_dataset=False, 
				   indicate_novel=False,
				   browser=False,
				   order='expression'):

		# check to see if input genes are in the graph
		if type(gids) != list:
			gids = [gids]
		for gid in gids:
			self.check_gene(gid)

		# check to see if these plotting settings will play together
		self.check_plotting_args(combine, indicate_dataset,
			indicate_novel, browser)

		# make sure all input datasets are present in graph
		if datasets == 'all':
			datasets = self.get_dataset_cols(include_annotation=False)
		elif not datasets:
			datasets = []
			order = 'tid'
		else:
			self.check_datasets(datasets)

		# now check to make sure abundance data is there for the
		# query columns, if user is asking
		if tpm or heatmap:
			self.check_abundances(datasets)
			tpm_cols = self.get_tpm_cols(datasets)
			report_cols = copy.deepcopy(tpm_cols)
		elif datasets:
			report_cols = datasets

		# order transcripts by user's preferences 
		self.order_transcripts(order)

		# loop through each gid and create the report
		for gid in gids:

			# subset based on gid
			t_df = self.t_df.loc[self.t_df.gid == gid].copy(deep=True)

			# if user doesn't care about datasets, just show all transcripts
			if not datasets:
				include_unexpressed = True
			# user only wants transcript isoforms that appear in their data
			if not include_unexpressed:
				counts_cols = self.get_count_cols(datasets)
				report_tids = t_df.loc[t_df[counts_cols].sum(axis=1)>0, 'tid']
			else:
				report_tids = t_df.loc[t_df.gid == gid, 'tid'].tolist()

			# plot each transcript with these settings
			self.plot_each_transcript(gid, prefix, combine,
									  indicate_dataset,
									  indicate_novel,
									  browser=browser)

			# if we're plotting tracks, we need a scale as well
			if not browser:
				report_type = 'swan'
			else:
				self.pg.plot_browser_scale()
				self.save_fig(prefix+'_browser_scale.png')
				report_type = 'browser'

			# if we're making a heatmap, need to first 
			# add pseudocount, log and normalize tpm values
			if heatmap:
				log_cols = ['{}_log_tpm'.format(d) for d in datasets]
				norm_log_cols = ['{}_norm_log_tpm'.format(d) for d in datasets]
				t_df[log_cols] = np.log2(t_df[tpm_cols]+1)
				max_val = max(t_df[log_cols].max().tolist())
				min_val = min(t_df[log_cols].min().tolist())
				t_df[norm_log_cols] = (t_df[log_cols]-min_val)/(max_val-min_val)
				report_cols = norm_log_cols

			# create report
			pdf_name = create_fname(prefix, 
						 combine,
						 indicate_dataset,
						 indicate_novel,
						 browser,
						 ftype='report',
						 gid=gid)
			report = Report(prefix, report_type, report_cols, datasets)
			report.add_page()

			# loop through each transcript and add it to the report
			for tid in report_tids:
				entry = t_df.loc[tid]
				## TODO would be faster if I didn't have to compute these names twice....
				## ie once in plot_each_transcript and once here
				fname = create_fname(prefix,
									 combine,
									 indicate_dataset,
									 indicate_novel, 
									 browser,
									 tid=entry.tid)
				report.add_transcript(entry, fname, heatmap=heatmap)
			report.write_pdf(pdf_name)

	##########################################################################
	############################# Error handling #############################
	##########################################################################

	# make sure that the set of arguments work with each other 
	# before we start plotting
	def check_plotting_args(self, combine, indicate_dataset, indicate_novel, browser=False):

		# can only do one or another
		if indicate_dataset and indicate_novel:
			raise Exception('Please choose either indicate_dataset '
							'or indicate_novel, not both.')

		# if indicate_dataset or indicate_novel are chosen, make sure
		# the dataset or annotation data exists in the SwanGraph
		if indicate_novel and 'annotation' not in self.get_dataset_cols():
			raise Exception('Annotation data not present in graph. Use  '
							'add_annotation before using indicate_novel')
		if indicate_dataset and indicate_dataset not in self.get_dataset_cols():
			raise Exception('Dataset {} not present in the graph. '
							''.format(indicate_dataset))

		# if browser, can't do indicate_novel, combine, or indicate_dataset
		if browser:
			if indicate_novel or indicate_dataset:
				raise Exception('Currently highlighting splice junctions in '
								'browser form is unsupported. Use browser option '
								'without inidicate_novel or inidicate_dataset.')
			if combine:
				raise Exception('Combining non-branching splice junctions '
								'not possible for browser track plotting.')


##########################################################################
################################## Extras ################################
##########################################################################








import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import copy
from collections import defaultdict
import matplotlib.patches as pch
from swan.utils import *
from swan.graph import Graph

class PlottedGraph(Graph):

	def __init__(self,
				 sg,
				 combine,
				 indicate_dataset,
				 indicate_novel,
				 tid=None,
				 browser=False,
				 gid=None):

		# save settings so we know how much we need to recompute
		self.combine = combine
		self.indicate_dataset = indicate_dataset 
		self.indicate_novel = indicate_novel
		# self.path = path
		self.gid = gid # if we're plotting a set of transcripts belonging to a specific gene
		               # or a summary graph of a specific gene
		self.tid = tid
		self.browser = browser

		# other fields to copy over
		self.datasets = sg.datasets

		# assign dfs for plotted graph
		self.G = sg.G.copy()
		self.loc_df = sg.loc_df.copy(deep=True)
		self.edge_df = sg.edge_df.copy(deep=True)
		self.t_df = sg.t_df.copy(deep=True)

		# have to deepcopy lists from t_df separately because of how pandas
		# handles lists
		self.t_df['path'] = self.t_df.apply(
			lambda x: copy.deepcopy(x.path), axis=1)

		if gid: 
			self.subset_on_gene(self.gid)
			self.g_min, self.g_max = self.get_gene_min_max(gid)
		if self.tid: 
			gid = self.get_gid_from_tid(self.tid)
			self.gid = gid
			self.subset_on_gene(gid)
			self.path = self.get_path_from_tid(self.tid)
		else:
			self.path = None

		## TODO is this the best way to handle browser vs. not browser graphs?
		# if we're making a swan graph
		if not browser:

			# used for getting plotting settings 
			self.loc_df['combined'] = False

			# combine non-branching paths
			if combine:
				nbps = self.find_nb_paths()
				self.agg_nb_nodes(nbps)

			# get positions/sizes of nodes, edges, and labels
			self.calc_pos_sizes()

			# get color/shape plotting settings for nodes, edges
			self.get_plt_settings()

		# if we're just making browser track things
		# elif browser:
		# 	1

	###############################################################################
	################### Getting plotting settings for nodes/edges #################
	###############################################################################

	# get color/shape plotting settings
	def get_plt_settings(self):

		# plotting styles
		gray = '#DCDCDC'
		yellow = '#F0E442'
		blue = '#0072B2'
		light_blue = '#56B4E9'
		red = '#D55E00'
		orange = '#E69F00'
		pink = '#CC79A7'
		green = '#009E73'

		color_dict = {'intron': pink, 
					  'exon': green,
					  'TSS': blue,
					  'alt_TSS': light_blue,
					  'TES': red,
					  'alt_TES': orange,
					  'internal': yellow,
					  'gray': gray}

		# get the list of datasets we should be looking at to determine
		# node or edge uniqueness
		dataset_cols = self.get_dataset_cols()

		# TODO use this when doing unique_to_dataset
		# if self.indicate_dataset:
		# 	dataset_cols.remove(self.indicate_dataset)

		###############
		#### NODES ####
		###############

		# node plotting settings: color
		self.loc_df = self.loc_df.apply(
			lambda x: self.get_node_color(x, color_dict), axis=1)

		# node plotting settings: shape
		self.loc_df['node_shape'] = self.loc_df.apply(
			lambda x: self.get_node_shape(x, dataset_cols), axis=1)

		###############
		#### EDGES ####
		###############

		# edge plotting settings: color
		self.edge_df['color'] = self.edge_df.apply(
			lambda x: self.get_edge_color(x, color_dict), axis=1)

		edge_dict = {'pos': 'arc3,rad=',
					 'neg': 'arc3,rad=-',
					 'straight': None}
		ordered_nodes = self.get_ordered_nodes()
		ordered_edges = [(n,m) for n,m in zip(ordered_nodes[:-1],ordered_nodes[1:])]

		# edge plotting settings: curve type 
		self.edge_df['raw_index'] = [i for i in range(len(self.edge_df.index))]
		self.edge_df['curve'] = self.edge_df.apply(
			lambda x: self.get_edge_style(x, ordered_edges, edge_dict), axis=1)
		self.edge_df.drop('raw_index', axis=1, inplace=True)

		# edge plotting settings: line type (dashed or not)
		self.edge_df['line'] = self.edge_df.apply(
			lambda x: self.get_edge_line(x, dataset_cols), axis=1)
		
	# calculates the positions and sizes of edges/nodes based on the 
	# number of nodes in the graph
	def calc_pos_sizes(self):

		pos = defaultdict()

		ordered_nodes = self.get_ordered_nodes()

		# check if forward or reverse strand 
		# is the first node in the ordered list a TSS or TES? 
		if self.loc_df.loc[ordered_nodes[0], 'TES']:
			ordered_nodes.reverse()

		y_coord = 0
		x_coords = np.linspace(-1, 1, len(ordered_nodes))
		for x_coord, n in zip(x_coords, ordered_nodes): 
			pos[n] = [x_coord, y_coord]

		x = len(ordered_nodes)

		# area-related sizes need to be non-linearly scaled
		# calculated by fitting power series curve to handpicked sizes
		node_size = 19248*(x**-1.14) 
		sub_node_size = node_size/2
		label_size = 43.9*(x**-0.484)

		# linearly-related sizes
		# rad_scale = ((11/540)*x)#+(73/540)
		edge_width = -x/18 + (121/18)

		# assign fields to plotted graph object 
		self.pos = pos
		self.node_size = node_size
		self.sub_node_size = sub_node_size
		self.label_size = label_size
		# self.rad_scale = rad_scale
		self.rad_scale = 0.35
		self.edge_width = edge_width

	# get the shape of the node 
	def get_node_shape(self, x, dataset_cols):

		# default shape
		shape = 'o'

		# hexagonal nodes for combined nodes 
		if x.combined:
			shape = 'H'
			# unique to dataset
			# if is_novel(self.indicate_novel, x) or unique_to_dataset(self.indicate_dataset, x, dataset_cols): 
			# 	shape = 'h'
			# in dataset  TODO make these separate options in the future
			if is_novel(self.indicate_novel, x) or in_dataset(self.indicate_dataset, x):
				shape = 'h'
		# diamond nodes for novel or inidicate_dataset nodes
		# if is_novel(self.indicate_novel, x) or unique_to_dataset(self.indicate_dataset, x, dataset_cols):
		if is_novel(self.indicate_novel, x) or in_dataset(self.indicate_dataset, x):
			shape = 'D'

		return shape

	# get the node color 
	def get_node_color(self, x, color_dict):

		# if we're dealing with a path, color only according
		# to the node's role in the path
		if self.tid:
			self.path = self.get_path_from_tid(self.tid)
			if x.vertex_id == self.path[0]:
				x['color'] = color_dict['alt_TSS']
				x['sub_color'] = color_dict['alt_TSS']
			elif x.vertex_id == self.path[-1]:
				x['color'] = color_dict['alt_TES']
				x['sub_color'] = color_dict['alt_TES']
			elif x.vertex_id in self.path:
				x['color'] = color_dict['internal']
				x['sub_color'] = color_dict['internal']	
			else:
				x['color'] = color_dict['gray']
				x['sub_color'] = color_dict['gray']	

		# combined nodes
		elif x.combined:
			types = x.combined_types

			# did one or two types of node go into this node?
			if len(types) == 2:

				x['sub_color'] = color_dict[types[0]]
				x['color'] = color_dict[types[1]]
			else: 
				x['sub_color'] = np.nan
				x['color'] = color_dict[types[0]]

		# non combined nodes
		else:

			# we don't need a sub color
			x['sub_color'] = np.nan

			# colors should label nodes by their 
			# MOST UNIQUE type (yes I know this is subjective) 
			# to me, this means that the label priority for a node is 
			# internal < TSS/alt_TSS < TES/alt_TES
			if x.internal: color = color_dict['internal']
			if x.TSS: color = color_dict['TSS']
			if x.alt_TSS: color = color_dict['alt_TSS']
			if x.TES: color = color_dict['TES']
			if x.alt_TES: color = color_dict['alt_TES']
			x['color'] = color

		return x

	# get the color of the edge
	def get_edge_color(self, x, color_dict):

		# firstly, if we're given a path, 
		# only color the edges that are in the path
		if self.path:
			path_edges = [(self.path[i],self.path[i+1])
						   for i in range(len(self.path)-1)]
			if x.edge_id in path_edges:
				color = color_dict[x.edge_type]
			else: color = color_dict['gray']

		else:
			color = color_dict[x.edge_type]

		return color

	# get the curve/style of the edge
	def get_edge_style(self, x, ordered_edges, edge_dict):

		# over 20 nodes, all should be curved
		if len(ordered_edges) < 20:
			if x.edge_id in ordered_edges:
				return edge_dict['straight']

		# make the arcs pretty
		## TODO maybe should have the arc height vary with number of nodes, ie 
		## shorter arc with less nodes? idk, still looks weird
		dist = self.pos[x.v2][0] - self.pos[x.v1][0]
		rad = self.rad_scale/dist

		if x.raw_index % 2 == 0:
			return edge_dict['pos']+str(rad)
		else:
			return edge_dict['neg']+str(rad)

	# get the style of the line for an edge
	def get_edge_line(self, x, dataset_cols):
		style = None
		# dashed if novel or unique to dataset
		# if is_novel(self.indicate_novel, x) or unique_to_dataset(self.indicate_dataset, x, dataset_cols):
		if is_novel(self.indicate_novel, x) or in_dataset(self.indicate_dataset, x):
			style = 'dashed'
		return style

	###############################################################################
	########################## Actually plotting stuff ############################
	###############################################################################

	# plots input plotted graph according to user's choices
	def plot_graph(self):

		# swan graph
		if not self.browser:
			self.plot_swan_graph()
		# browser track
		else:
			self.plot_browser_path()

	###############################################################################
	############################ Swan graph plotting ##############################
	###############################################################################

	# plots swan graph of current plotted graph
	def plot_swan_graph(self):

		# plotting stuff
		plt.figure(1, figsize=(14,2.8), frameon=False)
		plt.xlim(-1.05, 1.05)
		plt.ylim(-1.05, 1.05) 

		# plot edges, nodes, and labels
		self.plot_edges()

		# relabel nodes with strings for the purpose of labelling them
		relabel_map = {k:(k if type(k) == str else int(k)) for k in self.G.nodes}
		self.G = nx.relabel_nodes(self.G, relabel_map)
		nx.draw_networkx_labels(self.G,
			self.pos,
			font_size=self.label_size)

		self.plot_nodes()

	# plots edges from edge_df
	def plot_edges(self):
		for _, entry in self.edge_df.iterrows():
			edge = entry.edge_id
			nx.draw_networkx_edges(self.G, self.pos,
				edgelist=[edge],
				width=self.edge_width,
				edge_color=entry.color,
				connectionstyle=entry.curve,
				style=entry.line)

	# plots nodes from loc_df
	def plot_nodes(self):
		for _, entry in self.loc_df.iterrows():
			node = entry.vertex_id
			nx.draw_networkx_nodes(self.G, self.pos,
				nodelist=[node],
				node_color=entry.color,
				node_size=self.node_size,
				node_shape=entry.node_shape)
			if entry.combined:
				if len(entry.combined_types) == 2:
					nx.draw_networkx_nodes(self.G, self.pos,
						nodelist=[node],
						node_color=entry.sub_color,
						node_size=self.sub_node_size,
						node_shape=entry.node_shape)


	###############################################################################
	######################## Browser track style plotting #########################
	###############################################################################

	# plots the browser track representation of the transcript from self.path
	def plot_browser_path(self):

		# which gene does this transcript come from?
		# what are the min/max coords in the gene?
		gid = self.get_gid_from_tid(self.tid)
		g_min, g_max = self.get_gene_min_max(gid)
		t_min, t_max = self.get_transcript_min_max(self.tid)
		g_len = g_max - g_min

		x_min = int(g_min-(6/(g_max-g_min)))
		x_max = int(g_max+(6/(g_max-g_min)))

		strand = self.get_strand_from_tid(self.tid)
		path = self.get_path_from_tid(self.tid)

		# plotting init
		plt.figure(1, figsize=(14,2.8), frameon=False)
		plt.xlim(x_min, x_max)
		plt.ylim(-1.05, 1.05)
		ax = plt.gca()
		teal = '#014753'

		# plot each exon as a rectangle
		y_coord = -0.1
		height = 0.2
		exons = [(v1,v2) for v1,v2 in zip(path[:-1],path[1:])][::2]
		introns = [(v1,v2) for v1,v2 in zip(path[:-1],path[1:])][1::2]
		for v1,v2 in exons:
			x_coord = self.loc_df.loc[v1, 'coord']
			width = self.loc_df.loc[v2, 'coord'] - x_coord
			rect = pch.Rectangle((x_coord,y_coord), width, height, color=teal)
			ax.add_patch(rect)

		# plot each intron as a hashed line
		dist = 0.001*g_len
		plt.plot([t_min+dist, t_max-dist], [0,0], color=teal)

		# reverse plot if we're on the minus strand
		if strand == '-':
			plt.gca().invert_xaxis()

		def get_tick_coords(loc_df, gene_len, exons, g_min, g_max):

			tick_coords = list(np.linspace(g_min, g_max, 40))

			# remove ticks that are before the start of the first exon
			start = loc_df.loc[exons[0][1], 'coord']
			tick_coords = [t for t in tick_coords if t > start]

			# remove ticks that are after the end of the last exon
			stop = loc_df.loc[exons[-1][1], 'coord']
			tick_coords = [t for t in tick_coords if t < stop]

			# remove ticks in and around the area of plotted exons
			dist = 0.001*gene_len
			for v1,v2 in exons:
				start = loc_df.loc[v1, 'coord']
				stop = loc_df.loc[v2, 'coord']
				tick_coords = [t for t in tick_coords
							   if t < start-dist or t > stop+dist]

			# if we only have one intron, and nothing made the cut, just stick
			# a tick in the middle of the intron
			if len(exons) == 2 and not tick_coords:
				tick_coords = [(loc_df.loc[exons[0][1],'coord']+loc_df.loc[exons[1][0],'coord'])/2]

			return tick_coords 

		# get coordinates for evenly-spaced ticks indicating strandedness
		# ala genome browser
		tick_coords = get_tick_coords(self.loc_df, g_len, exons, g_min, g_max)
		plt.plot(tick_coords, [0 for i in range(len(tick_coords))],
			color=teal, marker='4')

	# plots the scale of a browser style graph
	def plot_browser_scale(self):

		# get plot limits
		g_min = self.g_min
		g_max = self.g_max 

		plt.figure(1, figsize=(14,1), frameon=False)
		plt.xlim(g_min, g_max)
		plt.ylim(-1.05, 1.05)
		ax = plt.gca()

		y_coord = -0.5
		height = 1
		gene_len = g_max - g_min
		width = int((10/10000)*gene_len)

		# try different scale lengths until we've reached something optimal
		scale_len = 500
		while 1:
			if scale_len >= gene_len:
				break
			scale_len *= 10

		# dial it back so we show a nice chunk instead of the scale of the entire gene
		scale_len /= 10

		# are we going to indicate this in bp or kb?
		if scale_len < 1000: 
			scale_unit = 'bp'
			text_scale = scale_len
		elif scale_len < 10**6:
			scale_unit = 'kb'
			text_scale =  scale_len / 1000
		else:
			scale_unit = 'mb'
			text_scale = scale_len / 10**6
		text_scale = int(text_scale)

		# center the edges of the scale bar
		gene_mid_pt = (g_max+g_min)/2
		scale_start = gene_mid_pt - (scale_len/2)
		scale_end = gene_mid_pt + (scale_len/2)

		# start bar
		rect = pch.Rectangle((scale_start, y_coord),
			width=width,
			height=height,
			color='k')
		ax.add_patch(rect)

		# end bar
		rect = pch.Rectangle((scale_end, y_coord),
			width=width,
			height=height,
			color='k')
		ax.add_patch(rect)

		# plot the line in between the bars
		plt.plot([scale_start+width, scale_end-width], [0,0],
			color='k',
			linewidth=2)

		# add scale text
		x_coord = scale_start - (gene_len/9)
		y_coord = -0.2
		txt = '{} {}'.format(text_scale, scale_unit)
		ax.text(x_coord, y_coord, txt, fontsize=30)

	###############################################################################
	####################### Combining non-branching paths #########################
	###############################################################################

	# starting nodes:
	# 1. have only one outgoing edge AND
	# 2. have more than one incoming edge | is a TSS | parent is TES AND
	# 3. node is not already in a nbp AND
	# 4. node is not a TES

	def is_nbp_start(self,n,nbps):
		nbp_nodes = [n for p in nbps for n in p]
		if self.G.out_degree(n) == 1: # 1
			tes_parent = False
			parents = list(self.G.predecessors(n))
			if parents: 
				for parent in parents:
					if self.loc_df.loc[parent, 'TES']:
						tes_parent = True
			if self.G.in_degree(n) >= 1 or self.loc_df.loc[n, 'TSS'] or tes_parent: # 2
				if n not in nbp_nodes and not self.loc_df.loc[n, 'TES']: # 3 & 4
					return True
		return False

	# end nodes:
	# 1. have more than one outgoing edge |
	# 2. have more than one incoming edge |
	# 3. is a TSS |
	# 4. parent node is a TES
	def is_nbp_end(self,n):
		if self.G.out_degree(n) > 1: # 1
			return True
		if self.G.in_degree(n) > 1: # 2
			return True
		if self.loc_df.loc[n, 'TSS']: # 3 
			return True
		if self.G.in_degree(n) == 1: # 4
			parent = list(self.G.predecessors(n))[0]
			if self.loc_df.loc[parent, 'TES']:
				return True
		return False

	# http://rosalind.info/problems/ba3m/ ty pavel 
	# modified to not group up TES/TSSs 
	# see is_nbp_start/end to see complete set of conditions
	def find_nb_paths(self):
		nbps = []
		# for v in self.G.nodes:
		for v in self.loc_df.vertex_id.tolist():
			if self.is_nbp_start(v,nbps):
				if self.G.out_degree(v) > 0:
					for w in self.G.successors(v):
						nbp = [v]
						while not self.is_nbp_end(w):
							nbp.append(w)
							succ = list(self.G.successors(w))
							if succ: w = succ[0]
							else: break
						if len(nbp) > 1: nbps.append(nbp)
		return nbps

	# aggregate nonbranching nodes and add to graph. remove old nodes
	# update loc_df, edge_df, and t_df to reflect these changes
	def agg_nb_nodes(self, nbps):

		G = self.G
		loc_df = self.loc_df
		edge_df = self.edge_df
		t_df = self.t_df

		# loop through each transcript path and each non-branching path,
		# replace nodes that will be aggregated 
		# need to use this because pandas copy DOES NOT deepcopy lists!
		t_df['new_path'] = t_df.apply(lambda x: copy.deepcopy(x.path), axis=1)
		paths = t_df.new_path.tolist() # this gives me a reference to the paths
									   # so when updating I'm directly modifying t_df
		for path in paths: 
			for combined_index, nbp in enumerate(nbps):

				# get the nodes that need to be deleted from the path
				# in the order they need to be deleted 
				del_nodes = sorted([int(i) for i in list(set(path) & set(nbp))],
					key=lambda x: loc_df.loc[x, 'coord'])

				# remove each node that is in a non-branching path
				for i, n in enumerate(del_nodes):
					if i == 0: 
						insertion_index = path.index(n)
					path.remove(n)
				if not del_nodes: insertion_index = -1

				# add aggregate node tuple instead
				if insertion_index != -1:
					path.insert(insertion_index, 'c{}'.format(combined_index))

		# shuffle around path columns in dataframe
		t_df.drop('path', axis=1, inplace=True)
		t_df.rename({'new_path': 'path'}, axis=1, inplace=True)

		mod_G = nx.DiGraph(G)

		combined_index = 0
		loc_df['agg_path'] = np.nan
		loc_df['combined'] = False
		loc_df['combined_types'] = np.nan

		# loop through each nonbranching path
		for path in nbps: 

			# some beginning of new loop things we need
			path = tuple([int(n) for n in path])
			start = path[0]
			stop = path[-1]
			combined_node = 'c{}'.format(combined_index)

			# get the colors for each aggregate node
			# TODO need to fix how these are colored with alt TSS/TES vs singleton TSS/TES
			combined_types = [j[0] for j in sorted([i for i in 
			   [('alt_TSS',loc_df.loc[path,'alt_TSS'].tolist().count(True)),
			   ('alt_TES',loc_df.loc[path,'alt_TES'].tolist().count(True)),
			   ('internal',loc_df.loc[path,'internal'].tolist().count(True))]
			   if i[1] != 0], key=lambda x: x[1], reverse=True)][:2]

			# compose the entries for edge_df for new edges that will be added
			# that include the aggregate node
			edge_attrs = {}
			if mod_G.in_degree(start) > 0: 
				in_nodes = list(mod_G.predecessors(start))
				old_edges = [(v1, start) for v1 in in_nodes]
				edge_attrs.update({(edge_id[0], combined_node): item for edge_id,item in edge_df.loc[old_edges,
					['strand', 'edge_type']].to_dict(orient='index').items()})
			if G.out_degree(stop) > 0: 
				out_nodes = list(mod_G.successors(stop))
				old_edges = [(stop, v2) for v2 in out_nodes]
				edge_attrs.update({(combined_node, edge_id[1]): item for edge_id,item in edge_df.loc[old_edges,
					['strand', 'edge_type']].to_dict(orient='index').items()})

			# compose the entry for loc_df for the aggregate node
			coord = loc_df.loc[start, 'coord']
			chrom = loc_df.loc[start, 'chrom']
			strand = loc_df.loc[start, 'strand']
			tss = alt_tss = tes = alt_tes = internal = False
			for n in path:
				if loc_df.loc[n, 'TSS']: tss = True
				if loc_df.loc[n, 'TES']: tes = True
				if loc_df.loc[n, 'alt_TSS']: alt_tss = True
				if loc_df.loc[n, 'alt_TES']: alt_tes = True
				if loc_df.loc[n, 'internal']: internal = True
			loc_attrs = {'chrom': chrom, 'strand': strand,
				   'vertex_id': combined_node,
				   'combined_types': combined_types,
				   'vertex_id_back': combined_node,
				   'agg_path': list(path), 'combined': True,
				   'TSS': tss, 'TES': tes,
				   'alt_TSS': alt_tss, 'alt_TES': alt_tes,
				   'internal': internal, 'coord': coord}

			# # if we're indicating dataset, which datasets are these nodes present in?
			# if args['indicate_dataset']:
			# 	node_attrs = assign_combined_datasets(mod_G, path, node_attrs)

			# add the new node and edges to graph
			mod_G.add_node(combined_node)
			for edge,item in edge_attrs.items():
				mod_G.add_edge(edge[0], edge[1])

			# also add aggregate node and all associated edges to loc_df, edge_df
			loc_df.reset_index(drop=True, inplace=True)
			edge_df.reset_index(drop=True, inplace=True)
			for edge, edge_attr in edge_attrs.items():
				edge_attr.update({'edge_id': edge, 'strand': strand,
								  'v1': edge[0], 'v2': edge[1]})
				edge_df = edge_df.append(edge_attr, ignore_index=True)
			loc_df = loc_df.append(loc_attrs, ignore_index=True)
			loc_df = create_dupe_index(loc_df, 'vertex_id')
			edge_df = create_dupe_index(edge_df, 'edge_id')
			loc_df = set_dupe_index(loc_df, 'vertex_id')
			edge_df = set_dupe_index(edge_df, 'edge_id')

			# remove all old nodes from the graph
			for n in path:
				mod_G.remove_node(n)

			# increment combined node index
			combined_index += 1

		# remove all old nodes/edges from loc_df and edge_df
		del_nodes = [n for path in nbps for n in path]
		del_edges = []
		del_edges += edge_df.loc[edge_df.v1.isin(del_nodes), 'edge_id'].tolist()
		del_edges += edge_df.loc[edge_df.v2.isin(del_nodes), 'edge_id'].tolist()
		del_edges = list(set(del_edges))

		loc_df = loc_df.loc[~loc_df.index.isin(del_nodes)] 
		edge_df = edge_df.loc[~edge_df.index.isin(del_edges)]

		self.G = mod_G
		self.loc_df = loc_df 
		self.edge_df = edge_df 
		self.t_df = t_df

	# gets nodes ordered by genomic position 
	def get_ordered_nodes(self):
		loc_ids = self.loc_df.vertex_id.tolist()
		coords = self.loc_df.coord.tolist()
		ordered_nodes = [i[0] for i in sorted(zip(loc_ids, coords),
			key=lambda x: x[1])]
		return ordered_nodes

# is this entry unique from all other datasets?
# and in the provided dataset?
def unique_to_dataset(dataset_name, x, dataset_cols):
	if not dataset_name: return False
	if x[dataset_name] and not any(x[dataset_cols].tolist()):
		return True
	return False

# is this entry in the provided dataset?
def in_dataset(dataset_name, x):
	if not dataset_name: return False
	return x[dataset_name]

# is this entry annotated?
def is_novel(indicate_novel, x):
	if not indicate_novel: return False
	if x.annotation: return False
	return True






from swan.utils import *
from fpdf import FPDF
import matplotlib.pyplot as pyplot

# report for genes - extension of FPDF class
class Report(FPDF):

	def __init__(self, prefix, report_type, report_cols, header_cols):
		super().__init__(orientation='L')

		# set report type, 'browser' or 'swan'
		if report_type != 'browser' and report_type != 'swan':
			raise Exception("Report type must be 'browser' or 'swan'.")
		else: 
			self.report_type = report_type

		# the columns that we'll include
		self.report_cols = report_cols
		self.header_cols = header_cols
		print('in report constructor')
		print('report cols:')
		print(self.report_cols)
		print()

		# prefix for files that we'll pull from 
		self.prefix = prefix

		# color map in case we're making a heatmap
		self.cmap = plt.get_cmap('Spectral_r')

	# header - should differ based on whether it's a browser report or
	# a swan report
	def header(self):
		self.set_font('Arial', 'B', 10)

		# add extra room for the scale if we're doing the browser version
		if self.report_type == 'swan':
			header_height = 10
		elif self.report_type == 'browser':
			header_height = 20

		self.cell(50, header_height, 'Transcript ID', border=True, align='C')
		for col in self.header_cols:
			self.cell(25, header_height, col,
					  border=True, align='C')

		# in case we need to add the browser models
		x = self.get_x()
		y = self.get_y()

		self.cell(100, header_height, 'Transcript Model', border=True, align='C')

		# add scale if we're doing browser models
		if self.report_type == 'browser':
			self.image(self.prefix+'_browser_scale.png',
					   x=x, y=y+12, w=100, h=50/7)
		self.ln()

	# add a transcript model to the report
	def add_transcript(self, entry, oname, heatmap=False):
		self.set_font('Arial', '', 10)
		self.cell(50, 20, entry['tid'], border=True, align='C')
		for col in self.report_cols:

			# heat map coloring
			if heatmap:
				color = self.cmap(entry[col])
				r = color[0]*255
				b = color[1]*255
				g = color[2]*255
				self.set_fill_color(r,b,g)
				border = False
				fill = True
				text = ''
			# TPM	
			elif '_tpm' in col:
				text = str(round(entry[col],2))
				border = True
				fill = False
			# presence/absence
			else:
				text = entry[col]
				if text == True:
					text = 'Yes'
				elif text == False:
					text = 'No'
				border = True
				fill = False 
			self.cell(25, 20, text, border=border, align='C', fill=fill)	
		x = self.get_x()
		y = self.get_y()

		# reset color to white
		self.set_fill_color(255,255,255)

		# embed transcript model
		self.cell(100, 20, '', border=True)
		self.image(oname, x=x, y=y, w=100, h=20)
		self.ln()


	# writes the pdf to file with the correct formatting
	def write_pdf(self, file):
		with open(file, 'wb') as outfile:
			outfile.write(self.output(dest='S').encode('latin-1'))


